# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import tensorflow as tf

if tf.__version__ < "2.6":
    from tensorflow.keras.layers import Conv2D, Dense, Reshape, ZeroPadding2D, AveragePooling2D, Activation, \
        DepthwiseConv2D, MaxPooling2D, ReLU, Add, Softmax, Concatenate, Multiply, Maximum, Minimum, BatchNormalization
else:
    from keras.layers import Conv2D, Dense, Reshape, ZeroPadding2D, AveragePooling2D, Activation, DepthwiseConv2D, \
        MaxPooling2D, ReLU, Add, Softmax, Concatenate, Multiply, Maximum, Minimum, BatchNormalization

from tensorflow.python.keras.layers.core import SlicingOpLambda
from tensorflow.python.ops.image_ops_impl import ResizeMethod
from model_compression_toolkit.core.common.target_platform.targetplatform2framework.attribute_filter import Eq

from model_compression_toolkit.core.tpc_models.tflite_tpc.v1.tp_model import get_tp_model
import model_compression_toolkit as mct

tp = mct.target_platform


def get_keras_tpc() -> tp.TargetPlatformCapabilities:
    """
    get a Keras TargetPlatformCapabilities object with default operation sets to layers mapping.
    Returns: a Keras TargetPlatformCapabilities object for the given TargetPlatformModel.
    """
    tflite_tp_model = get_tp_model()
    return generate_keras_tpc(name='tflite_keras', tp_model=tflite_tp_model)


def generate_keras_tpc(name: str, tp_model: tp.TargetPlatformModel):
    """
    Generates a TargetPlatformCapabilities object with default operation sets to layers mapping.

    Args:
        name: Name of the TargetPlatformCapabilities.
        tp_model: TargetPlatformModel object.

    Returns: a TargetPlatformCapabilities object for the given TargetPlatformModel.
    """

    keras_tpc = tp.TargetPlatformCapabilities(tp_model, name=name)

    with keras_tpc:
        tp.OperationsSetToLayers("NoQuantization", [AveragePooling2D,
                                                                tf.nn.avg_pool2d,
                                                                Concatenate,
                                                                tf.concat,
                                                                MaxPooling2D,
                                                                Multiply,
                                                                tf.multiply,
                                                                Reshape,
                                                                tf.reshape,
                                                                tp.LayerFilterParams(tf.image.resize,
                                                                                     method=ResizeMethod.BILINEAR),
                                                                tf.nn.space_to_depth,
                                                                ZeroPadding2D,
                                                                tf.gather,
                                                                tf.compat.v1.batch_to_space_nd,
                                                                tf.space_to_batch_nd,
                                                                tf.transpose,
                                                                tf.maximum,
                                                                Maximum,
                                                                tf.minimum,
                                                                Minimum,
                                                                tf.pad,
                                                                tf.slice,
                                                                SlicingOpLambda])

        tp.OperationsSetToLayers("FullyConnected", [Dense])
        tp.OperationsSetToLayers("L2Normalization", [tf.math.l2_normalize])
        tp.OperationsSetToLayers("LogSoftmax", [tf.nn.log_softmax])
        tp.OperationsSetToLayers("Tanh", [tf.nn.tanh, tp.LayerFilterParams(Activation, activation="tanh")])
        tp.OperationsSetToLayers("Softmax", [tf.nn.softmax,
                                             Softmax,
                                             tp.LayerFilterParams(Activation, activation="softmax")])
        tp.OperationsSetToLayers("Logistic", [tf.sigmoid, tp.LayerFilterParams(Activation, activation="sigmoid")])

        tp.OperationsSetToLayers("Conv2d", [Conv2D])
        tp.OperationsSetToLayers("DepthwiseConv2D", [DepthwiseConv2D])
        tp.OperationsSetToLayers("Relu", [tf.nn.relu,
                                          tf.nn.relu6,
                                          tp.LayerFilterParams(ReLU, Eq("max_value", None) | Eq("max_value", 6)),
                                          tp.LayerFilterParams(Activation, activation="relu")])
        tp.OperationsSetToLayers("Elu", [tf.nn.elu, tp.LayerFilterParams(Activation, activation="elu")])
        tp.OperationsSetToLayers("BatchNorm", [BatchNormalization, tf.nn.batch_normalization])
        tp.OperationsSetToLayers("Squeeze", [tf.squeeze])
        tp.OperationsSetToLayers("BiasAdd", [tf.nn.bias_add])
        tp.OperationsSetToLayers("Add", [tf.add, Add])

    return keras_tpc



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from model_compression_toolkit.core.common.target_platform.targetplatform2framework.current_tpc import  _current_tpc


class TargetPlatformCapabilitiesComponent:
    def __init__(self, name: str):
        self.name = name
        _current_tpc.get().append_component(self)

    def get_info(self):
        return {}



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Callable, List

from model_compression_toolkit.core import common
from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.core.common.constants import PYTORCH
from model_compression_toolkit.core.common.gptq.gptq_config import GradientPTQConfig
from model_compression_toolkit.core.common.target_platform import TargetPlatformCapabilities
from model_compression_toolkit.core.common.mixed_precision.kpi import KPI
from model_compression_toolkit.core.common.framework_info import FrameworkInfo
from model_compression_toolkit.core.common.network_editors.actions import EditRule
from model_compression_toolkit import CoreConfig, QuantizationConfig, DebugConfig
from model_compression_toolkit.core.common.mixed_precision.mixed_precision_quantization_config import \
    MixedPrecisionQuantizationConfig, DEFAULT_MIXEDPRECISION_CONFIG, MixedPrecisionQuantizationConfigV2
from model_compression_toolkit.core.common.post_training_quantization import post_training_quantization
from model_compression_toolkit.core.common.quantization.quantization_config import DEFAULTCONFIG

import importlib

if importlib.util.find_spec("torch") is not None:
    from model_compression_toolkit.core.pytorch.default_framework_info import DEFAULT_PYTORCH_INFO
    from model_compression_toolkit.core.pytorch.pytorch_implementation import PytorchImplementation
    from model_compression_toolkit.core.pytorch.constants import DEFAULT_TP_MODEL
    from torch.nn import Module

    from model_compression_toolkit import get_target_platform_capabilities
    DEFAULT_PYTORCH_TPC = get_target_platform_capabilities(PYTORCH, DEFAULT_TP_MODEL)


    def pytorch_gradient_post_training_quantization_experimental(in_module: Module,
                                                                 representative_data_gen: Callable,
                                                                 target_kpi: KPI = None,
                                                                 core_config: CoreConfig = CoreConfig(),
                                                                 fw_info: FrameworkInfo = DEFAULT_PYTORCH_INFO,
                                                                 gptq_config: GradientPTQConfig = None,
                                                                 target_platform_capabilities: TargetPlatformCapabilities = DEFAULT_PYTORCH_TPC):
        """
        Quantize a trained Pytorch module using post-training quantization.
        By default, the module is quantized using a symmetric constraint quantization thresholds
        (power of two) as defined in the default TargetPlatformCapabilities.
        The module is first optimized using several transformations (e.g. BatchNormalization folding to
        preceding layers). Then, using a given dataset, statistics (e.g. min/max, histogram, etc.) are
        being collected for each layer's output (and input, depends on the quantization configuration).
        Thresholds are then being calculated using the collected statistics and the module is quantized
        (both coefficients and activations by default).
        If gptq_config is passed, the quantized weights are optimized using gradient based post
        training quantization by comparing points between the float and quantized modules, and minimizing the
        observed loss.

        Args:
            in_module (Module): Pytorch module to quantize.
            representative_data_gen (Callable): Dataset used for calibration.
            target_kpi (KPI): KPI object to limit the search of the mixed-precision configuration as desired.
            core_config (CoreConfig): Configuration object containing parameters of how the model should be quantized, including mixed precision parameters.
            fw_info (FrameworkInfo): Information needed for quantization about the specific framework (e.g., kernel channels indices, groups of layers by how they should be quantized, etc.). `Default PyTorch info <https://github.com/sony/model_optimization/blob/main/model_compression_toolkit/core/pytorch/default_framework_info.py>`_
            gptq_config (GradientPTQConfig): Configuration for using gptq (e.g. optimizer).
            target_platform_capabilities (TargetPlatformCapabilities): TargetPlatformCapabilities to optimize the PyTorch model according to. `Default PyTorch TPC <https://github.com/sony/model_optimization/blob/main/model_compression_toolkit/core/tpc_models/pytorch_tp_models/pytorch_default.py>`_


        Returns:
            A quantized module and information the user may need to handle the quantized module.

        Examples:

            Import a Pytorch module:

            >>> import torchvision.models.mobilenet_v2 as models
            >>> module = models.mobilenet_v2()

            Create a random dataset generator:

            >>> import numpy as np
            >>> def repr_datagen(): return [np.random.random((1,224,224,3))]

            Import mct and pass the module with the representative dataset generator to get a quantized module:

            >>> import model_compression_toolkit as mct
            >>> quantized_module, quantization_info = mct.pytorch_post_training_quantization(module, repr_datagen)

        """

        if core_config.mixed_precision_enable:
            if not isinstance(core_config.mixed_precision_config, MixedPrecisionQuantizationConfigV2):
                common.Logger.error("Given quantization config to mixed-precision facade is not of type "
                                    "MixedPrecisionQuantizationConfigV2. Please use pytorch_post_training_quantization API,"
                                    "or pass a valid mixed precision configuration.")

            common.Logger.info("Using experimental mixed-precision quantization. "
                               "If you encounter an issue please file a bug.")

        return post_training_quantization(in_module,
                                          representative_data_gen,
                                          core_config,
                                          fw_info,
                                          PytorchImplementation(),
                                          target_platform_capabilities,
                                          gptq_config,
                                          target_kpi=target_kpi)


else:
    # If torch is not installed,
    # we raise an exception when trying to use these functions.
    def pytorch_gradient_post_training_quantization_experimental(*args, **kwargs):
        Logger.critical('Installing Pytorch is mandatory '
                        'when using pytorch_gradient_post_training_quantization_experimental. '
                        'Could not find the torch package.')



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import copy
from typing import Tuple, Callable

from model_compression_toolkit.core import common
from model_compression_toolkit.core.common.graph.base_graph import Graph
from model_compression_toolkit.core.common.graph.graph_matchers import EdgeMatcher, NodeOperationMatcher
from model_compression_toolkit.core.common.graph.base_node import BaseNode


class ResidualCollapsing(common.BaseSubstitution):
    """
    Collapse Add residual into previous Conv2D (No non-linear activation between them)
    """
    def __init__(self,
                 first_node: NodeOperationMatcher,
                 second_node: NodeOperationMatcher,
                 residual_collapsing_fn: Callable,
                 kernel_str: str,
                 layer_name_str: str = None):
        """
        Collapsing Add residual node into previous Conv2D node
        Args:
            first_node: Node matcher for convolution type nodes.
            second_node: Node matcher for add type nodes.
            residual_collapsing_fn: Function for updating the convolution kernel
            kernel_str: The framework specific attribute name of the convolution layer's weight/kernel.
            layer_name_str: The framework specific attribute name of layer's name.
        """
        super().__init__(matcher_instance=EdgeMatcher(first_node, second_node))
        self.residual_collapsing_fn = residual_collapsing_fn
        self.kernel_str = kernel_str
        self.layer_name_str = layer_name_str

    def substitute(self,
                   graph: Graph,
                   edge_nodes: Tuple[BaseNode, BaseNode]) -> Graph:
        """
        Collapse residual Add layer into previous Conv2D layers.
        Args:
            graph: Graph we apply the substitution on.
            edge_nodes: Tuple of two linear nodes
        Returns:
            Graph after applying the substitution.
        """

        first_node = edge_nodes[0]
        second_node = edge_nodes[1]

        # If the linear operator is part of a reused group (it is the "base" node, or a reused node),
        # we should skip the substitution.
        if first_node.reuse or first_node.reuse_group is not None:
            return graph
        if second_node.reuse or second_node.reuse_group is not None:
            return graph

        # Check if convolution and residual satisfy the collapsing conditions, otherwise skip substitution
        if len(graph.get_next_nodes(first_node)) > 1 or len(graph.get_prev_nodes(second_node)) != 2:
            return graph

        # Check if Add is residual connection, otherwise skip substitution
        conv_prev_nodes = graph.get_prev_nodes(first_node)
        add_prev_nodes = graph.get_prev_nodes(second_node)
        add_prev_nodes.remove(first_node)
        if conv_prev_nodes[0] != add_prev_nodes[0]:
            return graph

        # New collapsed weights
        kernel_collapsed = self.residual_collapsing_fn(first_node, self.kernel_str)

        num_nodes_before_substition = len(graph.nodes)
        num_edges_before_substition = len(graph.edges)

        # New collapsed node
        conv_collapsed = copy.deepcopy(first_node)
        conv_collapsed_name = first_node.name + '_' + second_node.name + "_collapsed"
        conv_collapsed.name = conv_collapsed_name
        conv_collapsed.set_weights_by_keys(self.kernel_str, kernel_collapsed)
        if self.layer_name_str is not None:
            conv_collapsed.framework_attr[self.layer_name_str] = conv_collapsed_name

        # Update graph
        graph.add_node(conv_collapsed)
        graph.reconnect_out_edges(current_node=second_node, new_node=conv_collapsed)
        graph.reconnect_in_edges(current_node=first_node, new_node=conv_collapsed)
        graph.replace_output_node(current_node=second_node, new_node=conv_collapsed)

        graph.remove_edge(first_node, second_node)
        graph.remove_edge(add_prev_nodes[0], second_node)
        graph.remove_node(first_node)
        graph.remove_node(second_node)

        # Sanity check
        assert num_nodes_before_substition - len(graph.nodes) == 1
        assert num_edges_before_substition - len(graph.edges) == 2

        return graph



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from tensorflow.keras.layers import DepthwiseConv2D, Conv2D, Conv2DTranspose, Dense
from tensorflow_model_optimization.python.core.quantization.keras.default_8bit.default_8bit_quantize_registry import \
    QuantizeConfig

from model_compression_toolkit.core import common
from model_compression_toolkit.core.common.framework_info import FrameworkInfo
from model_compression_toolkit.core.keras.quantizer.mixed_precision.selective_quantize_config import \
    SelectiveQuantizeConfig



def quantization_config_builder_mixed_precision(n: common.BaseNode,
                                                fw_info: FrameworkInfo) -> QuantizeConfig:
    """
    Build a QuantizeConfig for layers that should be wrapped in a QuantizeWrapper to
    be part of a mixed-precision model.
    Args:
        n: Node to build its QuantizeConfig.
        fw_info: Framework information (e.g., mapping from layers to their attributes to quantize).
    Returns:
        QuantizeConfig to wrap the layer so it can work in MP Keras models.
    """

    assert n.candidates_quantization_cfg is not None
    node_q_cfg_candidates = n.candidates_quantization_cfg
    # sort by decending bit width so using indices would be easier
    node_q_cfg_candidates.sort(key=lambda x: (x.weights_quantization_cfg.weights_n_bits,
                                              x.activation_quantization_cfg.activation_n_bits), reverse=True)

    float_weights = [n.get_weights_by_keys(attr) for attr in fw_info.get_kernel_op_attributes(n.type)]

    max_cfg_candidates = n.find_max_candidates_indices()
    assert len(max_cfg_candidates) == 1, \
        f"A maximal config candidate must be defined, but some node have multiple potential maximal candidates"
    max_candidate_idx = max_cfg_candidates[0]

    # Create a SelectiveQuantizeConfig that holds the float and quantized weights (every weight is
    # quantized using all possible bitwidhts in the node's candidates weights quantization configurations).
    return SelectiveQuantizeConfig(node_q_cfg=node_q_cfg_candidates,
                                   float_weights=float_weights,
                                   weight_attrs=fw_info.get_kernel_op_attributes(n.type),
                                   max_candidate_idx=max_candidate_idx)



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Tuple
import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import Conv2D
from model_compression_toolkit.core.common import BaseNode
from model_compression_toolkit.core.common.graph.graph_matchers import NodeOperationMatcher, NodeFrameworkAttrMatcher
from model_compression_toolkit.core.common.substitutions.linear_collapsing import Conv2DCollapsing
from model_compression_toolkit.core.keras.constants import KERNEL, KERNEL_SIZE, STRIDES, DILATIONS, LINEAR, \
    ACTIVATION, BIAS, USE_BIAS, LAYER_NAME, FILTERS, PADDING, GROUPS, DATA_FORMAT
from model_compression_toolkit.core.common.logger import Logger


def linear_collapsing_node_matchers() -> Tuple[NodeOperationMatcher, NodeOperationMatcher]:
    """
    Function generates matchers for matching:
    (Conv2D, Conv2D)[activation=linear] -> Conv2D.
    Returns:
        Matcher for 2 consecutive linear convolution
    """
    first_node = NodeOperationMatcher(Conv2D)
    second_node = NodeOperationMatcher(Conv2D)
    activation_linear = NodeFrameworkAttrMatcher(ACTIVATION, LINEAR)
    first_node = first_node & activation_linear
    return first_node, second_node


def conv2d_collapsing_fn(first_node: BaseNode,
                         second_node: BaseNode,
                         kernel_str: str,
                         bias_str: str) -> Tuple[np.ndarray, np.ndarray]:
    """
    Collapsing 2 convolutions to one convolution: Out = k2*(k1*x+b1)+b2 = k2*k1*x+k2*b1+b2 = k*x+b
    We calculate k=k2*k1 (collapsed kernel) by injecting identity tensor to the convolutions and extract the output
    We calculate b=k2*b1+b2 (collapsed bias) matrix multiplication
    Args:
        first_node: First layer node to collapse to second layer node
        second_node: Second layer node
        kernel_str: The framework specific attribute name of the convolution layer's weight/kernel.
        bias_str: The framework specific attribute name of the convolution layer's bias.
    Returns:
        The modified layer node's weights: kernel, bias
    """
    if first_node.type == Conv2D and second_node.type == Conv2D:
        # Get nodes attributes
        kernel1 = first_node.get_weights_by_keys(kernel_str)
        kernel2 = second_node.get_weights_by_keys(kernel_str)
        bias1 = first_node.get_weights_by_keys(bias_str)
        bias2 = second_node.get_weights_by_keys(bias_str)
        strides1 = first_node.framework_attr[STRIDES]
        strides2 = second_node.framework_attr[STRIDES]

        # --------------------------------------- #
        # Kernel collapsing: k=k2*k1
        # --------------------------------------- #
        # Inspired by https://arxiv.org/pdf/2103.09404.pdf - Algorithm1

        # Generate identity input with padding
        kx, ky = kernel1.shape[0] + kernel2.shape[0] - 1, kernel1.shape[1] + kernel2.shape[1] - 1
        x_pad, y_pad = 2 * kx - 1, 2 * ky - 1
        in_tensor = tf.eye(kernel1.shape[2])
        in_tensor = tf.expand_dims(tf.expand_dims(in_tensor, 1), 1)
        in_tensor = tf.pad(in_tensor, paddings=[[0, 0],
                                                [int(np.ceil((x_pad - 1) / 2)), int(np.floor((x_pad - 1) / 2))],
                                                [int(np.ceil((y_pad - 1) / 2)), int(np.floor((y_pad - 1) / 2))],
                                                [0, 0]])

        # Run first Conv2D
        conv1_out = tf.nn.conv2d(in_tensor, filters=kernel1, strides=strides1, padding="VALID")

        # Run second Conv2D
        conv2_out = tf.nn.conv2d(conv1_out, filters=kernel2, strides=strides2, padding="VALID")

        # Extract collapsed kernel from output: the collapsed kernel is the output of the convolution after fixing the dimension
        kernel_collapsed = tf.transpose(tf.reverse(conv2_out, [1, 2]), [1, 2, 0, 3]).numpy()

        # --------------------------------------- #
        # Bias collapsing: b=k2*b1+b2
        # --------------------------------------- #
        bias_collapsed = None
        if bias1 is not None:
            bias_collapsed = tf.squeeze(tf.matmul(bias1.reshape((1, -1)), tf.math.reduce_sum(kernel2, axis=(0, 1)))).numpy()
            if bias2 is not None:
                bias_collapsed += bias2
        elif bias2 is not None:
            bias_collapsed = bias2

        return kernel_collapsed, bias_collapsed
    else:
        Logger.error("No supported layer collapsing of {} and {}".format(first_node.type, second_node.type))


def keras_linear_collapsing() -> Conv2DCollapsing:
    """
    Returns:
        A Conv2DCollapsing initialized for Keras models.
    """
    first_node, second_node = linear_collapsing_node_matchers()
    return Conv2DCollapsing(first_node,
                            second_node,
                            conv2d_collapsing_fn,
                            KERNEL,
                            KERNEL_SIZE,
                            BIAS,
                            USE_BIAS,
                            STRIDES,
                            PADDING,
                            DILATIONS,
                            GROUPS,
                            FILTERS,
                            data_format_str=DATA_FORMAT,
                            layer_name_str=LAYER_NAME)



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from model_compression_toolkit.core import common
from model_compression_toolkit.core.common.collectors.statistics_collector import BaseStatsCollector


def create_stats_collector_for_node(node: common.BaseNode,
                                    output_channel_index: int) -> BaseStatsCollector:
    """
    Gets a node and a groups list and create and return a statistics collector for a node
    according to whether its statistics should be collected and the prior information we
    have about this node.

    Args:
        node: Node to create its statistics collector.
        output_channel_index: Index of output channels (for statistics per-channel).

    Returns:
        Statistics collector for statistics collection for the node.
    """

    if node.is_activation_quantization_enabled():
        stats_collector = common.StatsCollector(init_min_value=node.prior_info.min_output,
                                                init_max_value=node.prior_info.max_output,
                                                output_channel_index=output_channel_index)
    else:
        stats_collector = common.NoStatsCollector()

    return stats_collector



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from abc import ABC, abstractmethod
from collections import namedtuple

from model_compression_toolkit.core.common.graph.base_node import BaseNode
from model_compression_toolkit.core.common.quantization import quantization_params_generation
from model_compression_toolkit.core.common.quantization.quantization_params_fn_selection import \
    get_activation_quantization_params_fn, get_weights_quantization_params_fn

_EditRule = namedtuple('EditRule', 'filter action')


class EditRule(_EditRule):
    """
    A tuple of a node filter and an action. The filter matches nodes in the graph which represents the model,
    and the action is applied on these nodes during the quantization process.

    Examples:
        Create an EditRule to quantize all Conv2D wights using 9 bits:

        >>> import model_compression_toolkit as mct
        >>> from tensorflow.keras.layers import Conv2D
        >>> er_list = [EditRule(filter=mct.network_editor.NodeTypeFilter(Conv2D),
        >>> action=mct.network_editor.ChangeCandidatesWeightsQuantConfigAttr(weights_n_bits=9))]

        Then the rules list can be passed to :func:`~model_compression_toolkit.keras_post_training_quantization`
        to modify the network during the quantization process.

    """

    def __repr__(self):
        _str = f'filter={type(self.filter).__name__}{self.filter.__dict__}'
        _str = f'{_str} ; action={type(self.action).__name__}{self.action.kwargs}'
        return _str

    pass


class BaseAction(ABC):
    """
    Base class for actions. An action class applies a defined action on a node.
    """

    @abstractmethod
    def apply(self, node: BaseNode, graph, fw_info):
        """
        Apply an action on the node after matching the node with a node filter.

        Args:
            node: Node to apply the action on.
            graph: Graph to apply the action on.
            fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices,
                     groups of layers by how they should be quantized, etc.)

        Returns:
            Node after action is applied.

        """
        pass


class ChangeCandidatesWeightsQuantConfigAttr(BaseAction):
    """
    Change attributes in a layer's weights quantization configuration candidates.
    """

    def __init__(self, **kwargs):
        """
        Args:
            kwargs: Dictionary of attr_name and attr_value to change layer's weights quantization configuration candidates.
        """
        self.kwargs = kwargs

    def apply(self, node: BaseNode, graph, fw_info):
        """
        Change the attribute 'attr_name' in weights quantization config candidates with 'attr_value'.

        Args:
            node: Node object to change its quant_config.
            graph: Graph to apply the action on.
            fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices,
                     groups of layers by how they should be quantized, etc.)
        Returns:
            The node after its weights' quantization config candidates have been modified.
        """
        for nqc in node.candidates_quantization_cfg:
            for attr_name, attr_value in self.kwargs.items():
                nqc.weights_quantization_cfg.set_quant_config_attr(attr_name, attr_value)


class ChangeFinalWeightsQuantConfigAttr(BaseAction):
    """
    Change attributes in a layer's final weights quantization config.
    """

    def __init__(self, **kwargs):
        """
        Args:
            kwargs: Dictionary of attr_name and attr_value to change layer's final weights quantization config.
        """
        self.kwargs = kwargs

    def apply(self, node: BaseNode, graph, fw_info):
        if node.final_weights_quantization_cfg is not None:
            for attr_name, attr_value in self.kwargs.items():
                node.final_weights_quantization_cfg.set_quant_config_attr(attr_name, attr_value)


class ChangeCandidatesActivationQuantConfigAttr(BaseAction):
    """
    Change attributes in a layer's activation quantization configuration candidates.
    """

    def __init__(self, **kwargs):
        """
        Args:
            kwargs: Dictionary of attr_name and attr_value to change in the layer's activation quantization configuration candidates.
        """
        self.kwargs = kwargs

    def apply(self, node: BaseNode, graph, fw_info):
        """
        Change the attribute 'attr_name' in activation quantization configuration candidates with 'attr_value'.

        Args:
            node: Node object to change its quant_config.
            graph: Graph to apply the action on.
            fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices,
                     groups of layers by how they should be quantized, etc.)
        Returns:q
            The node after its activation quantization configuration candidates have been modified.
        """
        for nqc in node.candidates_quantization_cfg:
            for attr_name, attr_value in self.kwargs.items():
                nqc.activation_quantization_cfg.set_quant_config_attr(attr_name, attr_value)


class ChangeFinalActivationQuantConfigAttr(BaseAction):
    """
    Change attributes in a layer's final activation quantization config.
    """

    def __init__(self, **kwargs):
        """
        Args:
            kwargs: Dictionary of attr_name and attr_value to change layer's final activation quantization config.
        """
        self.kwargs = kwargs

    def apply(self, node: BaseNode, graph, fw_info):
        if node.final_activation_quantization_cfg is not None:
            for attr_name, attr_value in self.kwargs.items():
                node.final_activation_quantization_cfg.set_quant_config_attr(attr_name, attr_value)


class ChangeQuantizationParamFunction(BaseAction):
    """
    Class ChangeQuantizationParamFunction to change a node's weights/activations quantization params function.
    """

    def __init__(self, activation_quantization_params_fn=None, weights_quantization_params_fn=None):
        """
        Init a ChangeQuantizationParamFunction object.

        Args:
            activation_quantization_params_fn: a params function for a node's activations.
            weights_quantization_params_fn: a params function for a node's weights.
        """
        self.activation_quantization_params_fn = activation_quantization_params_fn
        self.weights_quantization_params_fn = weights_quantization_params_fn

    def apply(self, node: BaseNode, graph, fw_info):
        """
        Change the node's weights/activations quantization params function.

        Args:
            node: Node object to change its quantization params function.
            graph: Graph to apply the action on.
            fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices,
                     groups of layers by how they should be quantized, etc.)

        Returns:
            The node after its quantization params function has been modified.
        """
        for nqc in node.candidates_quantization_cfg:
            if self.activation_quantization_params_fn is not None:
                nqc.activation_quantization_cfg.set_activation_quantization_params_fn(
                    self.activation_quantization_params_fn)
            if self.weights_quantization_params_fn is not None:
                nqc.weights_quantization_cfg.set_weights_quantization_params_fn(self.weights_quantization_params_fn)


class ChangeCandidatesActivationQuantizationMethod(BaseAction):
    """
    Class ChangeQuantizationMethod to change a node's activations quantizer function.
    """

    def __init__(self, activation_quantization_method=None):
        """
        Init a ChangeCandidatesActivationQuantizationMethod object.

        Args:
            activation_quantization_method: a quantization method for a node's activations.
        """
        self.activation_quantization_method = activation_quantization_method

    def apply(self, node: BaseNode, graph, fw_info):
        """
        Change the node's activations quantization function.

        Args:
            node: Node object to change its threshold selection function.
            graph: Graph to apply the action on.
            fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices,
                     groups of layers by how they should be quantized, etc.)

        Returns:
            The node after its quantization function has been modified.
        """
        if self.activation_quantization_method is not None:
            for qc in node.candidates_quantization_cfg:
                activation_quantization_params_fn = get_activation_quantization_params_fn(
                    self.activation_quantization_method, qc.activation_quantization_cfg.activation_error_method)

                if node.prior_info.is_output_bounded():
                    activation_quantization_params_fn = quantization_params_generation.no_clipping_selection_min_max

                qc.activation_quantization_cfg.set_activation_quantization_params_fn(activation_quantization_params_fn)
                activation_quantization_fn = fw_info.activation_quantizer_mapping.get(self.activation_quantization_method)

                if activation_quantization_fn is None:
                    raise Exception('Unknown quantization method for activations')

                qc.activation_quantization_cfg.set_activation_quantization_fn(activation_quantization_fn)


class ChangeFinalWeightsQuantizationMethod(BaseAction):
    """
    Class ChangeFinalWeightsQuantizationMethod to change a node's weights/activations quantizer function.
    """

    def __init__(self, weights_quantization_method=None):
        """
        Init a ChangeFinalWeightsQuantizationMethod object.

        Args:
            weights_quantization_method: a quantization method for a node's weights.
        """

        self.weights_quantization_method = weights_quantization_method

    def apply(self, node: BaseNode, graph, fw_info):
        """
        Change the node's weights quantization function.

        Args:
            node: Node object to change its threshold selection function.
            graph: Graph to apply the action on.
            fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices,
                     groups of layers by how they should be quantized, etc.)

        Returns:
            The node after its quantization function has been modified.
        """

        if self.weights_quantization_method is not None:

            weights_quantization_params_fn = get_weights_quantization_params_fn(self.weights_quantization_method,
                                                                                node.final_weights_quantization_cfg.weights_threshold_method)

            node.final_weights_quantization_cfg.set_weights_quantization_params_fn(weights_quantization_params_fn)

            weights_quantization_fn = fw_info.weights_quantizer_mapping.get(self.weights_quantization_method)

            if weights_quantization_fn is None:
                raise Exception('Unknown quantization method for weights')

            node.final_weights_quantization_cfg.set_weights_quantization_fn(weights_quantization_fn)


class ChangeCandidatesWeightsQuantizationMethod(BaseAction):
    """
    Class ChangeCandidatesWeightsQuantizationMethod to change a node's weights quantizer function.
    """

    def __init__(self, weights_quantization_method=None):
        """
        Init a ChangeCandidatesWeightsQuantizationMethod object.

        Args:
            weights_quantization_method: a quantization method for a node's weights.
        """
        self.weights_quantization_method = weights_quantization_method

    def apply(self, node: BaseNode, graph, fw_info):
        """
        Change the node's weights quantization function.

        Args:
            node: Node object to change its threshold selection function.
            graph: Graph to apply the action on.
            fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices,
                     groups of layers by how they should be quantized, etc.)

        Returns:
            The node after its quantization function has been modified.
        """

        if self.weights_quantization_method is not None:
            for qc in node.candidates_quantization_cfg:

                weights_quantization_params_fn = get_weights_quantization_params_fn(self.weights_quantization_method,
                                                                                    qc.weights_quantization_cfg.weights_error_method)

                qc.weights_quantization_cfg.set_weights_quantization_params_fn(weights_quantization_params_fn)

                weights_quantization_fn = fw_info.weights_quantizer_mapping.get(self.weights_quantization_method)

                if weights_quantization_fn is None:
                    raise Exception('Unknown quantization method for weights')

                qc.weights_quantization_cfg.set_weights_quantization_fn(weights_quantization_fn)



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import copy

from model_compression_toolkit.core import common


def linear_collapsing_substitute(graph: common.Graph,
                                 linear_collapsing_substitution: common.BaseSubstitution) -> common.Graph:
    """
    Apply a list of linear collapsing substitutions on a graph.
    We run on the graph and find matches. For each valid match we do substitution.
    After each substitution we find matches again on the transformed graph to look for new matches.
    This is because a node can participate in more than one match so after substitution
    the matches are not valid anymore, and we can find new matches.
    Args:
        graph: Graph to transform.
        linear_collapsing_substitution: substitution to apply on the graph.
    Returns:
        Transformed graph after applying all linear collapsing substitutions.
    """
    graph = copy.deepcopy(graph)
    matched_nodes = graph.filter(linear_collapsing_substitution.matcher_instance)
    matched_nodes_list = []
    match_indicator = True
    while len(matched_nodes) > 0 and match_indicator:
        match_indicator = False
        for matched_node in matched_nodes:
            if matched_node not in matched_nodes_list:
                # Substitute
                graph = linear_collapsing_substitution.substitute(graph, matched_node)
                matched_nodes_list.append(matched_node)
                match_indicator = True
                break
        # Find new matches on the transformed graph
        matched_nodes = graph.filter(linear_collapsing_substitution.matcher_instance)
    return graph